from abc import ABC, abstractclassmethod
import numpy as np
from numpy.core.numeric import indices
from numpy.ma.core import concatenate

GRAVITATION_CONST = 6.76 * 10 ** -11

class AbsBodySystem(ABC):
    _dim: int
    _count: int
    _positions: np.ndarray
    _velocities: np.ndarray
    _accelerations: np.ndarray
    _masses: np.ndarray
    _dtype_out: callable

    @abstractclassmethod
    def simulate(self, dt: float):
        pass
    
    @abstractclassmethod
    def positions(self):
        pass

    @abstractclassmethod
    def velocities(self):
        pass

    @abstractclassmethod
    def accelerations(self):
        pass


class CpuBodySystem(AbsBodySystem):
    def __init__(self, bodies: list, res: np.ndarray=np.array([1000, 1000]), dim=2, dtype=int, scale=1.0) -> None:
        self._scale = scale
        self._count = len(bodies)
        self._dim = dim
        shape = (self._count, dim)
        self._masses = np.array(bodies, dtype=np.double)
        self._radii = (3 * self._masses / (4 * np.pi)) ** (1/3) * self._scale
        self._positions = np.random.random(shape) * res
        self._velocities = np.random.random(shape) / self._masses[:, None] * 1000
        self._accelerations = np.zeros(shape, dtype=np.double)
        self._dtype_out = dtype

    @property
    def positions(self):
        return self._positions.astype(self._dtype_out)

    @property
    def velocities(self):
        return self._velocities.astype(self._dtype_out)

    @property
    def accelerations(self):
        return self._accelerations.astype(self._dtype_out)

    @property
    def weights(self):
        return self._masses.astype(self._dtype_out)

    @property
    def radii(self):
        return self._radii.astype(self._dtype_out)


    def _handle_collisions(self):
        i = 0
        while i < len(self.positions):
            pos = self._positions[i]
            radius = self._radii[i]
            dif = self._positions - pos
            dist = np.sqrt(np.sum(dif**2, axis=1))

            # m1 * v1 + m2 * v2 = m1' * v1'
            do_collide = dist < (self._radii + radius)
            if np.count_nonzero(do_collide) > 1:
                positions  = self._positions[do_collide]
                velocities = self._velocities[do_collide]
                masses     = self._masses[do_collide]
                acc        = self._accelerations[do_collide][np.max(masses) == masses][0]
                impuls     = velocities * masses[:, None]
                mass       = np.sum(masses)
                vel        = np.sum(impuls, axis=0) / mass
                radius     = (3 * mass / (4 * np.pi)) ** (1/3) * self._scale


                self._positions     =   np.array([*self._positions      [do_collide == False], positions[np.max(masses) == masses][0]])
                self._velocities    =   np.array([*self._velocities     [do_collide == False], vel], dtype=float)
                self._accelerations =   np.array([*self._accelerations  [do_collide == False], acc])
                self._masses        =   np.array([*self._masses         [do_collide == False], mass])
                self._radii         =   np.array([*self._radii          [do_collide == False], radius])
            i+=1


    def simulate(self, dt: float= 1.0):
        #self._handle_collisions()
        for i, pos in enumerate(self._positions):
            dif = self._positions - pos
            dist = np.sqrt(np.sum(dif**2, axis=1))
            
            # Berechnung der Gravitationskraft
            dif = dif[dist != 0]
            dist = dist[dist != 0]

            accelerations = dif / dist[:, None] ** 3
            self._accelerations[i] += GRAVITATION_CONST * np.sum(accelerations, axis=0)

        # Bewegung

        self._velocities += self._accelerations * dt
        self._positions += self._velocities * dt
