from numba import cuda, types
import numpy as np
from time import time


MAX_DIMENSION = 3
TPB = 32 # Threads per block
MINMAL_DIST = 0.4


@cuda.jit
def calculate_acceleration_kernel(masses, positions, results):
    """
    int DIM >= 1
    masses: np.ndarray, shape=DIM
    positions: np.ndarray, shape(DIM, DIM)
    results: np.ndarray, shape(DIM, DIM, DIM)

    Creates an DIM x DIM gpu grid, where each thread calculates an single velocity.
    """
    # thread position in gpu grid
    tx, ty = cuda.threadIdx.x, cuda.threadIdx.y
    bx, by = cuda.blockIdx.x, cuda.blockIdx.y
    x, y = tx + bx * cuda.blockDim.x, ty + by * cuda.blockDim.y

    if x >= len(positions) or y >= len(positions):
        return

    # 
    DIM = positions.shape[1]
    p1 = positions[x]
    p2 = positions[y]
    m2 = masses[y]
    result = results[x]
    
    # calculate delta vector between p1 and p2
    delta_vector = cuda.local.array(MAX_DIMENSION, types.float64) # 3d
    for i in range(DIM):
        delta_vector[i] = p1[i] - p2[i]	

    # calculate distance between p1 and p2
    dist = 0
    for i in range(DIM):
        dist += delta_vector[i] ** 2
    dist = dist ** (1 / 2) + MINMAL_DIST

    # calculate force
    if dist == 0:
        acc = 0
    else:
        acc = -1 * m2 / dist ** 2 

    # save result in results
    for i in range(DIM):
        result[i][y] = acc * delta_vector[i]
    return


@cuda.jit
def sum_3d_array_on_axis_0(vector_field, results):
    """
    vector_field: np.ndarray, shape=(a, b, c)
    resuls: np.ndarray, shape=(a, b)

    Returns the sum of each component from each vector inside the given 2d vector field.
    """

    # get positions in gpu grid
    tx, ty = cuda.threadIdx.x, cuda.threadIdx.y
    bx, by = cuda.blockIdx.x, cuda.blockIdx.y
    x, y = tx + bx * cuda.blockDim.x, ty + by * cuda.blockDim.y

    width, height, depth = vector_field.shape[0], vector_field.shape[1], vector_field.shape[2]

    if x >= width or y >= height or depth == 0:
        return

    vector = vector_field[x][y]
    result = results[x]

    # sum components
    tmp = 0
    for i in range(depth):
        tmp += vector[i]
    result[y] += tmp
    
    return


def calculate_acceleration(masses, positions, velocities=None, dtype=np.float32):  
    """
    n = vector amount
    dim = vector height

    masses: np.ndarray, shape=(n)
    positions: np.ndarray, shape=(n, dim)
    velocities: np.ndarray, shape=(n, dim), default=np.full(0, dtype=np.float32)
    dtype: np.type, default=np.float32

    This function calculates the gravitational force for each mass on the given position.
    All computations are done on the gpu.
    """ 

    # converte to dtype
    masses = np.array(masses, dtype=dtype)
    positions = np.array(positions, dtype=dtype)
    velocities = np.array(velocities, dtype=dtype) if velocities is not None else np.full_like(positions, 0, dtype=dtype)

    # create result array
    dim = positions.shape[1]
    n = positions.shape[0]
    results_gpu = cuda.device_array((n, dim, n), dtype)

    # send data to gpu
    masses_gpu = masses if cuda.is_cuda_array(masses) else cuda.to_device(masses)
    positions_gpu = positions if cuda.is_cuda_array(positions) else cuda.to_device(positions)
    velocities_gpu = velocities if cuda.is_cuda_array(velocities) else cuda.to_device(velocities)

    # create gpu grid
    bpg = (positions.size + (TPB - 1)) // TPB

    # call gravity kernel
    calculate_acceleration_kernel[(bpg, bpg), (TPB, TPB)](masses_gpu, positions_gpu, results_gpu)

    # create gpu grid
    bpg = (n * dim + (TPB - 1)) // TPB

    # call merge kernel
    sum_3d_array_on_axis_0[(bpg, bpg), (TPB, TPB)](results_gpu, velocities_gpu)

    # get results back to cpu
    velocities = velocities_gpu.copy_to_host()
    return velocities


def calc_needed_bytes(count, dim, dtype=np.float32):
    positions = np.empty((count, dim), dtype=dtype)
    print(f"positions: {positions.nbytes / 10 ** 6} mb")
    velocities = np.empty_like(positions, dtype=dtype)
    print(f"velocities: {velocities.nbytes / 10 ** 6} mb")
    masses = np.empty(count, dtype=dtype)
    print(f"masses: {masses.nbytes / 10 ** 6} mb")
    results = np.empty((count, dim, count), dtype=dtype)
    print(f"results: {results.nbytes / 10 ** 6} mb")

if __name__ == "__main__":
    # positions = [[10], [0],[-10]]
    count = 100_000

    calc_needed_bytes(count, 2, dtype=np.float32)

    # # load data to gpu
    # positions = cuda.to_device(np.random.random((count, 2)))
    # masses = cuda.to_device([1] * len(positions))
    # avg = 0
    # steps = 10
    # for n in range(steps):
    #     t = time()
    #     velocities = calculate_acceleration(masses, positions, dtype=np.float32)
    #     dt = time() - t
    #     avg += dt

    # print(f"avg({count})={avg/steps}s")

