import numpy as np
from visualiser import PygameVisulizer, pygame
from nvidia_version.gravity_kernel import calculate_acceleration, cuda


SCALE = 1
SCREEN_SIZE = np.array([1000, 1000])
VIEW_OFFSET = np.zeros(2)
DT = 10
DICHTE = 1000

# Helper functions

def get_screen_position(position):
    """
    Center the position in the middle of the screen
    """
    return (position * SCALE) + SCREEN_SIZE / 2 + VIEW_OFFSET

def scale_radius(mass):
    return np.sqrt(mass * DICHTE / (2 * np.pi))

def scale_image(im, mass):
    scale = scale_radius(mass)
    scaled = pygame.transform.scale(im, (scale, scale))
    return scaled


# draw functions

def build_draw_planet_as_circle_function(masses, velocities, positions):
    def pygame_draw_planets(surface: pygame.Surface):
        global masses, velocities, positions

        velocities = calculate_acceleration(masses, positions, velocities)
        positions += velocities * DT * visulizer.dt

        max_mass = np.max(masses)
        for mass, planet_position in zip(masses, positions):
            color = (mass / max_mass * 200 + 55, 100, 100)
            pygame.draw.circle(surface, color, get_screen_position(planet_position[:2]), scale_radius(mass))

    return pygame_draw_planets


def build_draw_planet_as_image_function(masses, velocities, positions, alpha=80):
    # load image
    with open("planet.png") as f:
        im = pygame.image.load(f)

    im.set_alpha(alpha)

    def pygame_draw_planets(surface: pygame.Surface):
        global masses, velocities, positions

        velocities = calculate_acceleration(masses, positions, velocities)
        positions += velocities * DT * visulizer.dt

        for planet_position, mass in zip(positions, masses):
            image = scale_image(im, mass)
            im_rect = image.get_rect()
            center_offset = np.array([im_rect.width, im_rect.height]) / 2
            surface.blit(image, get_screen_position(planet_position[:2]) - center_offset)

        # debug circle
        # pygame.draw.circle(surface, 255, get_screen_position(positions[0]), 5)

    return pygame_draw_planets


def debug(event):
    global SCALE, VIEW_OFFSET, DT

    if event.type == pygame.KEYDOWN:
        # get positions
        if event.key == pygame.K_RETURN:
            print(" DEBUG ".center(20, "#"))
            print("Positions: ", positions)
        
        # restart
        if event.key == pygame.K_DELETE:
            init(n, dim)
        
        # dt
        if event.key == pygame.K_k:
            DT -= 1
            print("dt: ", DT)

        if event.key == pygame.K_j:
            DT += 1
            print("dt: ", DT)
    
    if event.type == pygame.MOUSEBUTTONDOWN:
        if event.button in [1, 2]:
            VIEW_OFFSET -= np.array(pygame.mouse.get_pos()) - SCREEN_SIZE / 2
            print(f"view_offset: {VIEW_OFFSET}")

        # increase scale
        if event.button == 4:
            SCALE *= 1.5
            print(f"scale: {SCALE}")

        if event.button == 5:
            SCALE /= 1.5
            print(f"scale: {SCALE}")
        
        
    return


def init(n, dim):
    global positions, velocities, masses

    # create arrays
    center_offset = np.array((1000, 1000))
    positions = np.random.randint(-center_offset, center_offset, (n, dim))
    velocities = np.full_like(positions, 0)
    masses = np.random.randint(1, 10, n)

    # load arrays explicit to gpu
    positions = cuda.to_device(positions)
    velocities = cuda.to_device(velocities)
    masses = cuda.to_device(masses)

    return


if __name__ == "__main__":
    # Konstanten
    n = 1_000
    dim = 2

    # init
    init(n, dim)
    draw = build_draw_planet_as_image_function(masses, positions, velocities)

    # main loop
    visulizer = PygameVisulizer(*SCREEN_SIZE)
    visulizer.register_draw_function(draw)
    visulizer.register_event_function(debug)
    visulizer.run()






